/*
 * Copyright © 2025 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of Nvidia Corporation and its affiliates
 * (the "Company") and all right, title, and interest in and to the software
 * product, including all associated intellectual property rights, are and
 * shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 */
#define _GNU_SOURCE

#include <inttypes.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "api/clx_api.h"
#include "api/clx_api_metrics.h"

#define RANDOM_SEED    12345
#define DATA_PAGE_SIZE 1024 * 64  // 64K

// Enum for metric indices (0-based)
enum MetricIndices {
    PHY_STATE_IDX = 0,
    LOGICAL_STATE_IDX = 1,
    PHY_MNGR_FSM_STATE_IDX = 2,
    LINK_SPEED_ACTIVE_IDX = 3,
    LINK_WIDTH_ACTIVE_IDX = 4
};

// Constants for link metric values (matching Lua script)
#define PHY_MNGR_FSM_STATE_ACTIVE   3
#define PHY_STATE_LINK_UP           5
#define LOGICAL_STATE_ACTIVE        4
#define LINK_SPEED_ACTIVE_NDR       8     // NDR speed state value
#define LINK_WIDTH_ACTIVE           4     // Active link width state value

// Number of test iterations
#define NUM_ITERATIONS 10000
// Probability (in percent) that a metric will change its state in each iteration
#define CHANGE_PROBABILITY 20
// Probability (in percent) that a metric will go to UP state when changing
#define UP_STATE_PROBABILITY 80
// Number of label sets (links) to simulate
#define NUM_LABEL_SETS 20

static void init_clx_api_params(clx_api_params_t* params) {
    params->buffer_size = DATA_PAGE_SIZE;

    // Enable Lua RTA exporter
    char* lua_script_dir = getenv("CLX_LUA_SCRIPT_DIR");
    if (lua_script_dir) {
        params->enabled_exporters = "rta_exporter";
    }
}

// Function to generate a random value within a range
static int random_range(int min, int max) { return min + rand() % (max - min + 1); }  // NOLINT

// Function to decide if an event should occur based on probability percentage
static bool should_occur(int probability_percent) { return (random_range(1, 100) <= probability_percent); }

// Function to check if all metrics are in their expected states for a given set
static bool are_metrics_up(int values[5]) {
    return (values[PHY_STATE_IDX] == PHY_STATE_LINK_UP &&
            values[LOGICAL_STATE_IDX] == LOGICAL_STATE_ACTIVE &&
            values[PHY_MNGR_FSM_STATE_IDX] == PHY_MNGR_FSM_STATE_ACTIVE &&
            values[LINK_SPEED_ACTIVE_IDX] == LINK_SPEED_ACTIVE_NDR &&
            values[LINK_WIDTH_ACTIVE_IDX] == LINK_WIDTH_ACTIVE);
}

// Function to generate a random initial value for a metric
static int get_random_initial_value(int metric_index) {
    switch (metric_index) {
        case PHY_STATE_IDX:  // phy_state metric
            return random_range(0, PHY_STATE_LINK_UP);
        case LOGICAL_STATE_IDX:  // logical_state metric
            return random_range(0, LOGICAL_STATE_ACTIVE);
        case PHY_MNGR_FSM_STATE_IDX:  // phy_mngr_fsm_state metric
            return random_range(0, PHY_MNGR_FSM_STATE_ACTIVE);
        case LINK_SPEED_ACTIVE_IDX:  // link_speed_active metric
            return random_range(0, LINK_SPEED_ACTIVE_NDR);
        case LINK_WIDTH_ACTIVE_IDX:  // link_width_active metric
            return random_range(0, LINK_WIDTH_ACTIVE);
        default:
            return 0;
    }
}

// Function to simulate our 5 metrics
static int simulate_link_metrics(clx_metrics_api_context_t* ctx, int iteration) {
    // Keep track of current values for each label set
    static int current_values[NUM_LABEL_SETS][5] = {{0}};
    static bool initialized                      = false;

    // Initialize values to random states on first call
    if (!initialized) {
        for (int set = 0; set < NUM_LABEL_SETS; set++) {
            for (int i = 0; i <= LINK_WIDTH_ACTIVE_IDX; i++) {
                current_values[set][i] = get_random_initial_value(i);
            }
        }
        initialized = true;

        // Print initial states
        printf("Initial states:\n");
        for (int set = 0; set < NUM_LABEL_SETS; set++) {
            printf("Link %d metrics: [%d, %d, %d, %d, %d]\n", set + 1, current_values[set][0], current_values[set][1],
                   current_values[set][2], current_values[set][3], current_values[set][4]);
        }
    }

    const char* metric_names[] = {"phy_state", "logical_state", "phy_mngr_fsm_state", "link_speed_active", "link_width_active"};

    // Add a label set
    const char* label_names[] = {"port", "guid"};
    size_t label_set_id       = clx_api_metrics_add_label_names(ctx, label_names, 2);
    if (label_set_id == CLX_METRICS_API_LABEL_SET_ID_IS_INVALID) {
        return -1;
    }

    uint64_t timestamp = clx_api_get_timestamp();

    // Check if all metrics are in their expected states before simulation
    bool all_up = true;
    for (int set = 0; set < NUM_LABEL_SETS; set++) {
        if (!are_metrics_up(current_values[set])) {
            all_up = false;
            break;
        }
    }
    if (all_up) {
        printf("Iteration %d: All metrics for all links are in their expected states!\n", iteration);
        for (int set = 0; set < NUM_LABEL_SETS; set++) {
            printf("Link %d metrics: [%d, %d, %d, %d, %d]\n", set + 1, current_values[set][0], current_values[set][1],
                   current_values[set][2], current_values[set][3], current_values[set][4]);
        }
    }

    // Simulate each label set (link)
    for (int set = 0; set < NUM_LABEL_SETS; set++) {
        // Create unique label values for this set
        char port[8], guid[32];
        snprintf(port, sizeof(port), "%d", set + 1);
        snprintf(guid, sizeof(guid), "0x%016lx", 0xABCD00000000L + set);
        const char* label_values[] = {port, guid};

        // Update each metric independently for this link
        for (int i = 0; i <= LINK_WIDTH_ACTIVE_IDX; i++) {
            // Add the metric with its current value
            clx_api_metrics_add_counter(ctx, timestamp, metric_names[i], current_values[set][i], label_set_id, label_values, 2);

            // After adding current value, decide if this metric should change state
            // Each link makes its own random decisions
            if (should_occur(CHANGE_PROBABILITY)) {
                // Decide if it should go to expected state or not - independent for each link
                bool go_up = should_occur(UP_STATE_PROBABILITY);

                switch (i) {
                    case PHY_STATE_IDX:  // phy_state metric
                        current_values[set][i] = go_up ? PHY_STATE_LINK_UP : random_range(0, 4);
                        break;
                    case LOGICAL_STATE_IDX:  // logical_state metric
                        current_values[set][i] = go_up ? LOGICAL_STATE_ACTIVE : random_range(0, 3);
                        break;
                    case PHY_MNGR_FSM_STATE_IDX:  // phy_mngr_fsm_state metric
                        current_values[set][i] = go_up ? PHY_MNGR_FSM_STATE_ACTIVE : random_range(0, 2);
                        break;
                    case LINK_SPEED_ACTIVE_IDX:  // link_speed_active metric
                        current_values[set][i] = go_up ? LINK_SPEED_ACTIVE_NDR : random_range(0, LINK_SPEED_ACTIVE_NDR - 1);
                        break;
                    case LINK_WIDTH_ACTIVE_IDX:  // link_width_active metric
                        current_values[set][i] = go_up ? LINK_WIDTH_ACTIVE : random_range(0, LINK_WIDTH_ACTIVE - 1);
                        break;
                }
            }
        }
    }

    // Flush metrics
    clx_api_metrics_flush(ctx);

    return 0;
}

static bool test_event_provider_initialize(clx_api_context_t* ctx, clx_api_provider_t* provider) {
    (void)ctx;
    (void)provider;
    return true;
}

static clx_api_provider_t* get_provider(void) {
    clx_api_provider_t* provider = calloc(1, sizeof(clx_api_provider_t));
    if (!provider) {
        return NULL;
    }

    clx_api_version_t ver = {{1, 0, 0}};
    provider->name        = strdup("link_state_test");
    if (!provider->name) {
        free(provider);
        return NULL;
    }
    provider->version    = ver;
    provider->initialize = &test_event_provider_initialize;

    return provider;
}

int main(void) {
    // Set fixed random seed for reproducibility
    srand(RANDOM_SEED);

    // Initialize API parameters
    clx_api_params_t params = {0};
    init_clx_api_params(&params);

    // Create provider
    clx_api_provider_t* provider = get_provider();
    if (!provider) {
        fprintf(stderr, "Failed to initialize provider\n");
        return -1;
    }

    // Create contexts
    clx_api_context_t* clx_api_ctx = clx_api_create_context(&params, provider);
    clx_metrics_api_context_t* ctx = clx_api_metrics_create_context(clx_api_ctx);

    printf("Starting link state simulation for %d links...\n", NUM_LABEL_SETS);
    printf("Will run for %d iterations (duration will vary due to random delays 0-100ms)\n", NUM_ITERATIONS);

    // Run simulation for specified number of iterations
    for (int i = 0; i < NUM_ITERATIONS; i++) {
        simulate_link_metrics(ctx, i);
        usleep(random_range(0, 100) * 1000);  // Random sleep between 0-100ms (converted to microseconds)
    }

    // Cleanup
    clx_api_metrics_destroy_context(ctx);
    clx_api_destroy_context(clx_api_ctx);
    free(provider->name);
    free(provider);

    return 0;
}
